/* This file is part of the KDE libraries
   Copyright 2007 Matthew Woehlke <mw_triad@users.sourceforge.net>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License version 2, as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#include <math.h>

#include "kcolor.h"

#define CH_0 m_channels[0]
#define CH_1 m_channels[1]
#define CH_2 m_channels[2]
#define CH_3 m_channels[3]
#define CH_4 m_channels[4]
#define CH_5 m_channels[5]
#define CH_6 m_channels[6]
#define CH_7 m_channels[7]

#define CH_A CH_7

#define CH_R CH_0
#define CH_G CH_1
#define CH_B CH_2

#define CH_H CH_0
#define CH_S CH_1
#define CH_I CH_2

#define CH_V CH_I
#define CH_L CH_I
#define CH_Y CH_I

// What coefficients for RGB->Y shall we use?
#define HSY_REC 601 // use 601 for now
#if   HSY_REC == 601
static const double yCoefficient[] = { 0.299, 0.587, 0.114 };
#elif HSY_REC == 709
static const double yCoefficient[] = { 0.2126, 0.7152, 0.0722 };
#else // use Qt values
static const double yCoefficient[] = { 0.34375, 0.5, 0.15625 };
#endif

static inline double normalize( double a ) {
    return ( a > 0.0 ? ( a < 1.0 ? a : 1.0 ) : 0.0 );
}

static inline double wrap( double a, double d = 1.0 ) {
    register double r = fmod( a, d );
    return ( r < 0.0 ? d + r : ( r > 0.0 ? r : 0.0 ) );
}

static inline double mix( double a, double b, double k ) {
    return a + ( ( b - a ) * k );
}

/******************************************************************************
 * Constructors
 */
KColor::KColor() : m_space(Rgb), d(NULL)
{
    CH_R = 0.0;
    CH_G = 0.0;
    CH_B = 0.0;
    CH_A = 0.0;
}

KColor::KColor( ColorSpec space,
                double c0, double c1, double c2, double c3,
                double c4, double c5, double c6, double c7 )
    : m_space(space), d(NULL)
{
    m_channels[0] = c0;
    m_channels[1] = c1;
    m_channels[2] = c2;
    m_channels[3] = c3;
    m_channels[4] = c4;
    m_channels[5] = c5;
    m_channels[6] = c6;
    m_channels[7] = c7;
}

KColor::KColor( const QColor& color ) : m_space(Rgb), d(NULL)
{
    CH_R = color.redF();
    CH_G = color.greenF();
    CH_B = color.blueF();
    CH_A = color.alphaF();
}

KColor::KColor( const KColor& color ) : m_space( color.m_space ), d(NULL)
{
    m_channels[0] = color.m_channels[0];
    m_channels[1] = color.m_channels[1];
    m_channels[2] = color.m_channels[2];
    m_channels[3] = color.m_channels[3];
    m_channels[4] = color.m_channels[4];
    m_channels[5] = color.m_channels[5];
    m_channels[6] = color.m_channels[6];
    m_channels[7] = color.m_channels[7];
}

KColor::~KColor()
{
//     delete d; KColorPrivate not used currently
}

KColor& KColor::operator=( const KColor& color )
{
    m_space = color.m_space;
    m_channels[0] = color.m_channels[0];
    m_channels[1] = color.m_channels[1];
    m_channels[2] = color.m_channels[2];
    m_channels[3] = color.m_channels[3];
    m_channels[4] = color.m_channels[4];
    m_channels[5] = color.m_channels[5];
    m_channels[6] = color.m_channels[6];
    m_channels[7] = color.m_channels[7];
    return *this;
}

KColor& KColor::operator=( const QColor& color )
{
    m_space = Rgb;
    CH_R = color.redF();
    CH_G = color.greenF();
    CH_B = color.blueF();
    CH_A = color.alphaF();
    return *this;
}

KColor KColor::fromRgb( double r, double g, double b, double a )
{
    return KColor( Rgb, r, g, b, 0.0, 0.0, 0.0, 0.0, a );
}

KColor KColor::fromHsv( double h, double s, double v, double a )
{
    return KColor( Hsv, h, s, v, 0.0, 0.0, 0.0, 0.0, a );
}

KColor KColor::fromHsl( double h, double s, double l, double a )
{
    return KColor( Hsl, h, s, l, 0.0, 0.0, 0.0, 0.0, a );
}

KColor KColor::fromHsy( double h, double s, double y, double a )
{
    return KColor( Hsy, h, s, y, 0.0, 0.0, 0.0, 0.0, a );
}

/******************************************************************************
 * Colorspace Conversion
 */

static inline KColor rgb2hsl( double r, double g, double b, double a )
{
    double min;
    double max;
    double h, l, s;

    if ( r > g ) {
        max = ( r > b ? r : b );
        min = ( g < b ? g : b );
    } else {
        max = ( g > b ? g : b );
        min = ( r < b ? r : b );
    }

    l = 0.5 * (max + min);
    s = 0.0;
    h = 0.0;

    if ( max != min ) {
        double delta = max - min;

        if ( l <= 0.5 )
            s = delta / ( max + min );
        else
            s = delta / ( 2.0 - max - min );

        if ( r == max )
            h = ( g - b ) / delta;
        else if ( g == max )
            h = 2.0 + ( b - r ) / delta;
        else if ( b == max )
            h = 4.0 + ( r - g ) / delta;

        h /= 6.0;
        if ( h < 0.0 )
            h += 1.0;
    }

    return KColor::fromHsl( h, s, l, a );
}

static inline double h2c( double h, double m1, double m2 )
{
    h = wrap( h, 6.0 );

    if ( h < 1.0 )
        return mix( m1, m2, h );
    if ( h < 3.0 )
        return m2;
    if ( h < 4.0 )
        return mix( m1, m2, 4.0 - h );
    return m1;
}

static inline KColor hsl2rgb( double h, double s, double l, double a )
{
    double m1, m2;
    double r, g, b;

    // TODO h2rgb( h, r, g, b );
    h *= 6.0;

    if ( l <= 0.5 )
        m2 = l * ( 1.0 + s );
    else
        m2 = l + s * ( 1.0 - l );
    m1 = 2.0 * l - m2;

    r = h2c( h + 2.0, m1, m2 );
    g = h2c( h,       m1, m2 );
    b = h2c( h - 2.0, m1, m2 );

    return KColor::fromRgb( r, g, b, a );
}

KColor KColor::convert( const KColor& color, ColorSpec space )
{
    return color.convert( space );
}

KColor KColor::convert( ColorSpec space ) const
{
    // Do nothing if already in the desired space
    if ( space == m_space )
        return *this;

    if ( space == Rgb ) {
        // Convert to RGB
        switch ( m_space )
        {
            case Hsv:
                return KColor(); // TODO hsv2rgb( CH_H, CH_S, CH_V, CH_A );
            case Hsl:
                return hsl2rgb( CH_H, CH_S, CH_L, CH_A );
            case Hsy:
                return KColor(); // TODO hsy2rgb( CH_H, CH_S, CH_Y, CH_A );
            default: // should never get here!
                return KColor();
        }
    }
    else {
        // Do all conversion to/from Rgb, it's slower but less complicated
        if ( m_space != Rgb )
            return convert( Rgb ).convert( space );

        // Convert from Rgb
        switch ( space ) {
            case Hsv:
                return KColor(); // TODO rgb2hsv( CH_R, CH_G, CH_B, CH_A );
            case Hsl:
                return rgb2hsl( CH_R, CH_G, CH_B, CH_A );
            case Hsy:
                return KColor(); // TODO rgb2hsy( CH_R, CH_G, CH_B, CH_A );
            default: // should never get here!
                return KColor();
        }
    }
}

KColor::operator QColor()
{
    KColor c = convert( Rgb );
    return QColor::fromRgbF( normalize( c.CH_R ), normalize( c.CH_G ),
                             normalize( c.CH_B ), normalize( c.CH_A ) );
}

/******************************************************************************
 * Channel Accessors
 */

KColor::ColorSpec KColor::space() const
{
    return m_space;
}

double KColor::alpha()
{
    return CH_A;
}

#define CHGET( s, c ) \
{ \
    if ( m_space == s ) \
        return c; \
    else \
        return convert( s ).c; \
}

double KColor::red() const          CHGET( Rgb, CH_R )
double KColor::green() const        CHGET( Rgb, CH_G )
double KColor::blue() const         CHGET( Rgb, CH_B )

double KColor::value() const        CHGET( Hsv, CH_V )
double KColor::luminance() const    CHGET( Hsl, CH_L )
double KColor::luma() const         CHGET( Hsy, CH_Y )

double KColor::hue() const
{
    switch ( m_space ) {
        case Hsv:
        case Hsl:
        case Hsy:
            return CH_H;
        default:
            return convert( Hsl ).CH_H;
    }
}

double KColor::saturation( ColorSpec space ) const  CHGET( space, CH_S )

double KColor::luma( const QColor& color )
{
    return color.redF()   * yCoefficient[0]
         + color.greenF() * yCoefficient[1]
         + color.blueF()  * yCoefficient[2];
}

double KColor::luma( const KColor& color )
{
    return color.luma();
}

/******************************************************************************
 * Convenience wrappers
 */
KColor KColor::blend( const KColor& c1, const KColor& c2,
                      double k, double r, BlendMode mode, ColorSpec space )
{
    return c1.blend( c2, k, r, mode, space );
}

KColor KColor::blend( const QColor& c1, const QColor& c2,
                      double k, double r, BlendMode mode, ColorSpec space )
{
    return KColor( c1 ).blend( c2, k, r, mode, space );
}

KColor KColor::blend( const QColor& with, double k, double r,
                      BlendMode mode, ColorSpec space ) const
{
    return blend( KColor( with ), k, r, mode, space );
}

KColor KColor::darken( const KColor& color, double k,
                       ColorSpec space, double r )
{
    return color.darken( k, space, r );
}

KColor KColor::darken( const QColor& color, double k,
                       ColorSpec space, double r )
{
    return KColor( color ).darken( k, space, r );
}

KColor KColor::lighten( const KColor& color, double k,
                        ColorSpec space, double r )
{
    return color.lighten( k, space, r );
}

KColor KColor::lighten( const QColor& color, double k,
                        ColorSpec space, double r )
{
    return KColor( color ).lighten( k, space, r );
}

/******************************************************************************
 * Blending functions
 */

KColor KColor::darken( double k, ColorSpec space, double r ) const
{
    switch ( space ) {
        case Hsv:
        case Hsl:
        case Hsy:
        {
            KColor c = convert( space );
            c.CH_S *= r;
            c.CH_V *= k;
            return c;
        }
        default:
            return blend( fromRgb( 0.0, 0.0, 0.0 ), 1.0 - k, 0.5, BlendNormal, space );
    }
}

KColor KColor::lighten( double k, ColorSpec space, double r ) const
{
    switch ( space ) {
        case Hsv:
        case Hsl:
        case Hsy:
        {
            KColor c = convert( space );
            c.CH_S = 1.0 - normalize( ( 1.0 - c.CH_S ) * r );
            c.CH_V = 1.0 - normalize( ( 1.0 - c.CH_V ) * k );
            return c;
        }
        default:
            return blend( fromRgb( 1.0, 1.0, 1.0 ), 1.0 - k, 0.5, BlendNormal, space );
    }
}

KColor KColor::blend( const KColor& with, double k, double r,
                      BlendMode mode, ColorSpec space ) const
{
    const KColor c1 = convert( space ), c2 = with.convert( space );
    switch ( mode ) {
        case BlendAlpha: // SRC*(1-DST_A)*2*K + DST*DST_A*2*R
            switch ( space ) {
                case Hsv:
                case Hsl:
                case Hsy:
                    return KColor( space,
                                   mix( c1.CH_H, c2.CH_H, c2.CH_A ),
                                   mix( c1.CH_S, c2.CH_S, c2.CH_A ),
                                   mix( c1.CH_I * k, c2.CH_I * r, c2.CH_A ) * 2.0,
                                   0.0, 0.0, 0.0, 0.0,
                                   c1.CH_A );
                case Rgb:
                    return fromRgb( mix( c1.CH_R * k, c2.CH_R * r, c2.CH_A ) * 2.0,
                                    mix( c1.CH_G * k, c2.CH_G * r, c2.CH_A ) * 2.0,
                                    mix( c1.CH_B * k, c2.CH_B * r, c2.CH_A ) * 2.0,
                                    mix( c1.CH_A, c2.CH_A, k ) );
                default:
                    return KColor( space,
                                   mix( c1.CH_0 * k, c2.CH_0 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_1 * k, c2.CH_1 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_2 * k, c2.CH_2 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_3 * k, c2.CH_3 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_4 * k, c2.CH_4 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_5 * k, c2.CH_5 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_6 * k, c2.CH_6 * r, c2.CH_A ) * 2.0,
                                   mix( c1.CH_A, c2.CH_A, k ) );
            }
            break;
//         case BlendAdd: TODO
//         case BlendMultiply: TODO
        default: // BlendNormal: SRC*K*2*R + DST*Ki*2*R
            r *= 2.0;
            switch ( space ) {
                case Hsv:
                case Hsl:
                case Hsy:
                    return KColor( space,
                                   mix( c1.CH_H, c2.CH_H, k ),
                                   mix( c1.CH_S, c2.CH_S, k ),
                                   mix( c1.CH_I, c2.CH_I, k ) * r,
                                   0.0, 0.0, 0.0, 0.0,
                                   mix( c1.CH_A, c2.CH_A, k ) );
                case Rgb:
                    return fromRgb( mix( c1.CH_R, c2.CH_R, k ) * r,
                                    mix( c1.CH_G, c2.CH_G, k ) * r,
                                    mix( c1.CH_B, c2.CH_B, k ) * r,
                                    mix( c1.CH_A, c2.CH_A, k ) );
                default:
                    return KColor( space,
                                   mix( c1.CH_0, c2.CH_0, k ) * r,
                                   mix( c1.CH_1, c2.CH_1, k ) * r,
                                   mix( c1.CH_2, c2.CH_2, k ) * r,
                                   mix( c1.CH_3, c2.CH_3, k ) * r,
                                   mix( c1.CH_4, c2.CH_4, k ) * r,
                                   mix( c1.CH_5, c2.CH_5, k ) * r,
                                   mix( c1.CH_6, c2.CH_6, k ) * r,
                                   mix( c1.CH_A, c2.CH_A, k ) );
            }
    }
    // gcc seems to be laboring under the delusion that this next line of code
    // is actually reachable...
    return KColor();
}

// kate: space-indent on; indent-width 4; replace-tabs on; auto-insert-doxygen on;
