/* This file is part of the KDE libraries
   Copyright 2007 Matthew Woehlke <mw_triad@users.sourceforge.net>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License version 2, as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.
*/

#ifndef KCOLOR_H
#define KCOLOR_H

#include <QtGui/QColor>

class KColorPrivate;

class KColor
{
public:
    typedef enum {
        Rgb,            ///< Red, Green, Blue
        Hsv,            ///< Hue, Saturation, Value
        Hsl,            ///< Hue, Saturation, Luminance
        Hsy,            ///< Hue, Saturation, Luma (percieved brightness)
//         Cmyk,           ///< Cyan, Magenta, Yellow, Black -- NOT SUPPORTED YET
    } ColorSpec;

    typedef enum {
        BlendNormal,    ///< SRC*K*2*R + DST*Ki*2*R
        BlendAlpha,     ///< SRC*(1-DST_A)*2*K + DST*DST_A*2*R
        BlendAdd,       ///< SRC*K + DST*R
        BlendMultiply,  ///< SRC*(Ki + DST*K*R)
    } BlendMode;

    ///////////////////////////////////////////////////////////////////////////
    // Constructors
    ///////////////////////////////////////////////////////////////////////////

    explicit KColor();
    explicit KColor( const QColor& );
    KColor( const KColor& );
    virtual ~KColor();

    operator QColor();
    KColor& operator=( const QColor& );
    KColor& operator=( const KColor& );

    static KColor fromRgb( double red, double green, double blue, double alpha = 1.0 );
    static KColor fromHsv( double hue, double saturation, double value, double alpha = 1.0 );
    static KColor fromHsl( double hue, double saturation, double luminance, double alpha = 1.0 );
    static KColor fromHsy( double hue, double saturation, double luma, double alpha = 1.0 );

    ///////////////////////////////////////////////////////////////////////////
    // Channel Accessors
    ///////////////////////////////////////////////////////////////////////////

    KColor convert( ColorSpec space ) const;
    static KColor convert( const KColor& color, ColorSpec space );

    ColorSpec space() const;

    double alpha();

    double red() const;
    double green() const;
    double blue() const;

    double hue() const;
    double saturation( ColorSpec space ) const;
    double value() const;
    double luminance() const;
    double luma() const;

    static double luma( const QColor& );
    static double luma( const KColor& );

    // NOT SUPPORTED YET
//     double cyan();
//     double magenta();
//     double yellow();
//     double black();

    // Please read TODO for things that KColor "should" do but won't until
    // someone wants to use the functionality...

    ///////////////////////////////////////////////////////////////////////////
    // Blending Functions
    ///////////////////////////////////////////////////////////////////////////

    static KColor blend( const KColor& c1, const KColor& c2,
                         double k = 0.5, double r = 0.5,
                         BlendMode mode = BlendNormal,
                         ColorSpec space = Rgb );
    static KColor blend( const QColor& c1, const QColor& c2,
                         double k = 0.5, double r = 0.5,
                         BlendMode mode = BlendNormal,
                         ColorSpec space = Rgb );
    KColor blend( const KColor& with, double k = 0.5, double r = 0.5,
                  BlendMode mode = BlendNormal, ColorSpec space = Rgb ) const;
    KColor blend( const QColor& with, double k = 0.5, double r = 0.5,
                  BlendMode mode = BlendNormal, ColorSpec space = Rgb ) const;

    static KColor darken( const KColor& color, double k = 0.5,
                          ColorSpec space = Hsy, double r = 1.0 );
    static KColor darken( const QColor& color, double k = 0.5,
                          ColorSpec space = Hsy, double r = 1.0 );
    static KColor lighten( const KColor& color, double k = 0.5,
                           ColorSpec space = Hsy, double r = 1.0 );
    static KColor lighten( const QColor& color, double k = 0.5,
                           ColorSpec space = Hsy, double r = 1.0 );
    KColor darken( double k = 0.5, ColorSpec space = Hsy, double r = 1.0 ) const;
    KColor lighten( double k = 0.5, ColorSpec space = Hsy, double r = 1.0 ) const;

private:
    KColor( ColorSpec space,
            double c0, double c1, double c2, double c3,
            double c4, double c5, double c6, double c7 );

    ColorSpec m_space;
    double m_channels[8];
    KColorPrivate *d;
};

#endif // KCOLOR_H

// kate: space-indent on; indent-width 4; replace-tabs on; auto-insert-doxygen on;
